// Demonstrate Simple Texturing
// Based on checker.c from the OpenGL RedBook, first edition
// and on NeHe Lesson 6
//
// Ported to C++ and modified for Core Profile by Alex Clarke.

#include "Angel.h"
#include "IMGLoader.h"


//Vertex array objects
GLuint vao;

//Shader Programs
GLuint program1;

//Color Uniform for Shader
GLint shaderColor;


//Transformation variables
mat4 p;
mat4 mv;
GLuint projIndex;
GLuint mvIndex;

//Animation variables
GLuint lastUpdateTime;
GLfloat xrot, yrot, zrot;       // variables for x, y, and z rotation
GLint anim = 0;                 // animation toggle variable
GLfloat degreesPerSecond = 60;  // degrees per second base rotation rate

//Texture variables
GLuint texture[1];              // Texture names array (only one for now)


// Load Bitmaps And Convert To Textures
// Connect texture units to the shader
void ConfigureGLTextures()
{
    // Load Texture
    
    //Load picture from file
    Image image;
    if (!image.Load("Data/NeHe.png"))
    {
        exit(1);
    }
    
    // Create Texture Name and Bind it as current
    glGenTextures(1, &texture[0]);
    glBindTexture(GL_TEXTURE_2D, texture[0]);   // 2d texture (x and y size)
    
    //Set Texture Parameters
    // scale linearly when image bigger than texture
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
    // scale linearly when image smaller than texture
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
    
    glPixelStorei(GL_UNPACK_ALIGNMENT, 4);
    // Load texture into OpenGL RC
    glTexImage2D(GL_TEXTURE_2D,     // 2D texture
                 0,                 // level of detail 0 (full or no mipmap)
                 GL_RGB,	        // internal format - how GL will store tex.
                 image.sizeX,       // x size from image
                 image.sizeY,       // y size from image
                 0,	                // border 0 - only legal value in Core Profile
                 GL_BGR,            // format of image data in memory
                 GL_UNSIGNED_BYTE,  // data type of image data in memory
                 image.data         // image data itself
                 );
    
    //Once you load a texture is the original data still needed?
    //What happens if you overwrite it as below?
    //for (int i = 0; i < image.sizeX; i++)
    //	for (int j=0; j < image.sizeY*3; j++)
    //		image.data[i*image.sizeY*3+j] = 1;
    
    //If it isn't needed then is it OK to deallocate it?
    //delete []image.data;
    //image.data = 0;
    
    //Connect texture units to samplers inside the shader.
    glUniform1i(glGetUniformLocation(program1, "tex"), 0);
    
};


vec4 cubeVerts[] =
{
    vec4( 0.5, 0.5, 0.5, 1), //0
    vec4( 0.5, 0.5,-0.5, 1), //1
    vec4( 0.5,-0.5, 0.5, 1), //2
    vec4( 0.5,-0.5,-0.5, 1), //3
    vec4(-0.5, 0.5, 0.5, 1), //4
    vec4(-0.5, 0.5,-0.5, 1), //5
    vec4(-0.5,-0.5, 0.5, 1), //6
    vec4(-0.5,-0.5,-0.5, 1)  //7
};

vec4 points[] =
{
#define solidCubeStart 0
#define solidCubeVertices 36
    //Solid Cube - 36 vertices := 6 faces * 2 Triangles per face * 3 vertices per triangle
    cubeVerts[0], cubeVerts[4], cubeVerts[6],  //front
    cubeVerts[0], cubeVerts[6], cubeVerts[2],
    cubeVerts[1], cubeVerts[0], cubeVerts[2],  //right
    cubeVerts[1], cubeVerts[2], cubeVerts[3],
    cubeVerts[5], cubeVerts[1], cubeVerts[3],  //back
    cubeVerts[5], cubeVerts[3], cubeVerts[7],
    cubeVerts[4], cubeVerts[5], cubeVerts[7],  //left
    cubeVerts[4], cubeVerts[7], cubeVerts[6],
    cubeVerts[4], cubeVerts[0], cubeVerts[1],  //top
    cubeVerts[4], cubeVerts[1], cubeVerts[5],
    cubeVerts[6], cubeVerts[7], cubeVerts[3],  //bottom
    cubeVerts[6], cubeVerts[3], cubeVerts[2],
};


vec4 nRight =	vec4( 1.0f, 0.0f, 0.0f, 0.0f);
vec4 nLeft =		vec4(-1.0f, 0.0f, 0.0f, 0.0f);
vec4 top =		vec4( 0.0f, 1.0f, 0.0f, 0.0f);
vec4 bottom =	vec4( 0.0f,-1.0f, 0.0f, 0.0f);
vec4 front =	vec4( 0.0f, 0.0f, 1.0f, 0.0f);
vec4 back =		vec4( 0.0f, 0.0f,-1.0f, 0.0f);

vec4 white =	vec4( 1.0f, 1.0f, 1.0f, 1.0f);

vec4 normals[] =
{
    //Normals for Solid Cube
    front, front, front, front, front, front,
    nRight, nRight, nRight, nRight, nRight, nRight,
    back, back, back, back, back, back,
    nLeft, nLeft, nLeft, nLeft, nLeft, nLeft,
    top, top, top, top, top, top,
    bottom, bottom, bottom, bottom, bottom, bottom,
};


GLfloat texCoords[] =
{
    //Texture Coordinates for Solid Cube
    //Note that each face is the same.
    1, 1,	0, 1,	0, 0, // triangle 1
    1, 1,	0, 0,	1, 0, // triangle 2
    
    1, 1,	0, 1,	0, 0, // triangle 1
    1, 1,	0, 0,	1, 0, // triangle 2
    
    1, 1,	0, 1,	0, 0, // triangle 1
    1, 1,	0, 0,	1, 0, // triangle 2
    
    1, 1,	0, 1,	0, 0, // triangle 1
    1, 1,	0, 0,	1, 0, // triangle 2
    
    1, 1,	0, 1,	0, 0, // triangle 1
    1, 1,	0, 0,	1, 0, // triangle 2
    
    1, 1,	0, 1,	0, 0, // triangle 1
    1, 1,	0, 0,	1, 0, // triangle 2
};



//----------------------------------------------------------------------------

// OpenGL initialization
void init()
{
    glClearColor(0.5, 0.5, 0.5, 1.0);
    
    //Set up shaders and VAOs
    program1 = InitShader( "vShaderL5-2.glsl", "fShaderL5-2.glsl" );
    
    
    glUseProgram( program1 );
    glGenVertexArrays( 1, &vao );
    glBindVertexArray( vao );
    
    //Load vertex positions and texture coordinates into vertex buffer.
    GLuint buffer;
    glGenBuffers( 1, &buffer );
    glBindBuffer( GL_ARRAY_BUFFER, buffer );
    
    //allocate memory for and store points, texture coordinates and normals.
    glBufferData( GL_ARRAY_BUFFER, sizeof(points)+sizeof(texCoords)+sizeof(normals), NULL, GL_STATIC_DRAW );
    glBufferSubData( GL_ARRAY_BUFFER, 0, sizeof(points), points );
    glBufferSubData( GL_ARRAY_BUFFER, sizeof(points), sizeof(texCoords), texCoords );
    glBufferSubData( GL_ARRAY_BUFFER, sizeof(points)+sizeof(texCoords), sizeof(normals), normals );
    
    //bind points, texture coordinates and normals to shader variables
    GLuint vPosition = glGetAttribLocation( program1, "vPosition" );
    glEnableVertexAttribArray( vPosition );
    glVertexAttribPointer( vPosition, 4, GL_FLOAT, GL_FALSE, 0, BUFFER_OFFSET(0) );
    GLuint vTexCoord = glGetAttribLocation( program1, "vTexCoord" );
    glEnableVertexAttribArray( vTexCoord );
    glVertexAttribPointer( vTexCoord, 2, GL_FLOAT, GL_FALSE, 0, BUFFER_OFFSET(sizeof(points)) );
    GLuint vNormal = glGetAttribLocation( program1, "vNormal" );
    glEnableVertexAttribArray( vNormal );
    glVertexAttribPointer( vNormal, 4, GL_FLOAT, GL_FALSE, 0, BUFFER_OFFSET(sizeof(points)+sizeof(texCoords)) );
    
    //Configure some texture names and connect texture units to the shader
    ConfigureGLTextures();
    
    //Get transformation matrix locations from shader
    projIndex = glGetUniformLocation(program1, "p");
    mvIndex = glGetUniformLocation(program1, "mv");
    
    glEnable(GL_DEPTH_TEST);
}

//----------------------------------------------------------------------------

void display( void )
{
    //Clear the drawing buffer
    glClear( GL_COLOR_BUFFER_BIT |GL_DEPTH_BUFFER_BIT );
    
    mv = Translate(0.0f, 0.0f, -2.5f);
    
    // Clear The Screen And The Depth Buffer
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    
    mv *= RotateX(xrot);
    mv *= RotateY(yrot);
    mv *= RotateZ(zrot);
    
    //Select a texture unit
    glActiveTexture(GL_TEXTURE0);
    //Choose the texture to attach to the texture unit
    glBindTexture(GL_TEXTURE_2D, texture[0]);
    
    //Draw textured geometry
    glUniformMatrix4fv(mvIndex, 1, GL_TRUE, mv);
    glDrawArrays(GL_TRIANGLES, solidCubeStart, solidCubeVertices);
    
    //Show what was drawn
    glutSwapBuffers();
}

//----------------------------------------------------------------------------

void reshape (int w, int h)
{
    glViewport(0, 0, (GLsizei) w, (GLsizei) h);
    p = Perspective(45.0f,(GLfloat)w/h,0.1f,100.0f);
    glUniformMatrix4fv(projIndex, 1, GL_TRUE, p);
}


void animate()
{
    GLint currentTime = glutGet(GLUT_ELAPSED_TIME);
    GLint elapsed_ms = currentTime - lastUpdateTime;
    lastUpdateTime = currentTime;
    if (anim)
    {
        GLfloat baseRot = degreesPerSecond*elapsed_ms/1000.0f;
        xrot += baseRot; // Increase X Axis Rotation
        yrot += baseRot; // Increase Y Axis Rotation
        zrot += baseRot; // Increase Z Axis Rotation
    }
    glutPostRedisplay();
}

void keyboard(unsigned char key, int x, int y)
{
    switch (key)
    {
        case 'a':
        case 'A':
            // The letter "a" toggles animation
            lastUpdateTime = glutGet(GLUT_ELAPSED_TIME);
            anim = !anim;
            break;
        case 27:
        case 'q':
        case 'Q':
            exit(EXIT_SUCCESS);
    }
}

//----------------------------------------------------------------------------


int main( int argc, char **argv )
{
    glutInit( &argc, argv );
    glutInitWindowSize( 512, 512 );
    
#ifdef __APPLE__
    glutInitDisplayMode( GLUT_3_2_CORE_PROFILE | GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH );
#else
    // If you are using freeglut, the next two lines will check if
    // the code is truly 3.2. Otherwise, comment them out
    glutInitContextVersion( 3, 2 );
    glutInitContextProfile( GLUT_CORE_PROFILE );
    glutInitDisplayMode( GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH );
    
#endif
    
    glutCreateWindow( "Lab 5 - Textured Cube" );
    glewExperimental = GL_TRUE;
    glewInit();
    
    init();
    
    glutIdleFunc(animate);
    glutDisplayFunc( display );
    glutKeyboardFunc( keyboard );
    glutReshapeFunc( reshape );
    glutMainLoop();
    return 0;
}
